// Imperative and Object-Oriented Programming Methodology
// Uppsala University - 2018 Autumn
// Copyright © 2018 Henrik Schulze & Tobias Wrigstad
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include "yourcode.h"

struct list {
  link_t *sentinel;
  link_t *last;
  void_cmp cmp;
};

static link_t *link_create(link_t *next, void *element) {
  link_t *link = malloc(sizeof(link_t));
  link->next = next;
  link->element = element;
  return link;
}

static void link_destroy(link_t *remove_me) {
  free(remove_me->element);
  free(remove_me);
}

list_t *list_create(void_cmp cmp) {
  list_t *list = malloc(sizeof(list_t));
  list->sentinel = list->last = link_create(NULL, NULL); /// Dummy!
  list->cmp = cmp;
  return list;
}

void list_insert(list_t *list, void *element) {
  link_t *cursor = list->sentinel;
  while (cursor->next && list->cmp(cursor->next->element, element) < 0) {
    cursor = cursor->next;
  }
  if (cursor->next) {
    if (list->cmp(cursor->next->element, element)) {
      cursor->next = link_create(cursor->next, element);
    } else {
      // cursor->next->element = element;
    }
  } else {
    list->last = cursor->next = link_create(NULL, element);
  }
}

int list_size(list_t *list) {
  /// TODO: Returnera antalet element i listan. Detta skall räknas ut
  /// genom iteration över samtliga länkar.
  int size = 0;
  for (link_t *cursor = list->sentinel->next; cursor; cursor = cursor->next) {
    ++size;
  }
  return size;
}

void list_destroy(list_t *list) {
  /// TODO: Ta bort listan, alla länkar och alla länkars alla element
  // if (list==NULL) { return; }
  if (list!=NULL) {
    link_t *prev = list->sentinel;
    link_t *crsr = list->sentinel->next;
    while (crsr!=NULL) {
      link_destroy(prev);
      prev = crsr;
      crsr = crsr->next;
    }
    link_destroy(prev);
  }
  free(list);
}

void list_merge(list_t *source, list_t *dest) {
  /// TODO: Alla länkar (och dess medföljande element, dock ej dummies)
  /// skall flyttas från source till dest.
  link_t *elem = source->sentinel->next;
  link_t *nextelem = elem!=NULL? elem->next : NULL;
  link_t *prev = dest->sentinel;
  link_t *crsr = dest->sentinel->next;
  while (crsr!=NULL) {
    // list_insert(dest, elem->element);
    // elem=elem->next;
    while (elem && dest->cmp(elem->element, crsr->element) < 0) {
      nextelem = elem->next;
      prev->next = elem;
      elem->next = crsr;
      prev = elem;
      elem = nextelem;
    }
    prev = crsr;
    crsr = crsr->next;
  }
  prev->next = elem; // Maybe NULL.
  while (elem != NULL) {
    prev = elem;
    elem = elem->next;
  }
  dest->last = prev;
  source->sentinel->next = NULL;
}

/// Returns a string representation of the links to be printed.
char *links_print(list_t *list, elem_to_string to_string) {
  char *result = calloc(2048, sizeof(char));
  strcat(result, "[");
  for (link_t *cursor = list->sentinel->next; cursor; cursor = cursor->next) {
    char *tmp = NULL;
    asprintf(&tmp, "%p, ", to_string ? to_string(cursor) : (char*)cursor);
    strcat(result, tmp);
    free(tmp);
  }
  int size = strlen(result);
  strcpy(result + (size > 1 ? size - 2 : size), "]");
  return result;
}

/// Returns a string representation of the list to be printed
/// No need to read or change this function!!
char *list_print(list_t *list, elem_to_string to_string) {
  char *result = calloc(2048, sizeof(char));
  strcat(result, "[");
  for (link_t *cursor = list->sentinel->next; cursor; cursor = cursor->next) {
    char *tmp = NULL;
    asprintf(&tmp, "%s, ", to_string ? to_string(cursor->element)
             : (char *)cursor->element);
    strcat(result, tmp);
    free(tmp);
  }
  int size = strlen(result);
  strcpy(result + (size > 1 ? size - 2 : size), "]");
  return result;
}

/// Return the sentinel link in the list for testing
/// No need to read or change this function!!
link_t *list_first(list_t *list) {
  return list->sentinel->next;
}
