#include <string.h>
#include "yourcode.h"

/// strdup equivalent since it does not always work on some machines
char *str(char *in)
{
  int len = strlen(in);
  char *out = calloc(len + 1, sizeof(char));
  for (int i = 0; i < len; ++i)
    {
      out[i] = in[i];
    }
  return out;
}

void create_random_lists(list_t **list1, list_t **list2)
{
  char *strings[] = {"a", "b", "c", "d", "e", "f", "g", "h",
                     "i", "j", "k", "l", "m", "n", "o", "p",
                     "q", "r", "s", "t", "u", "v", "w", "x",
                     "y", "z"};
  int nr_strings = 26;

  *list1 = list_create((void_cmp) strcmp);
  for (size_t input_size = 5 + (rand() % 5); input_size; --input_size)
    {
      int item = rand() % nr_strings; 
      list_insert(*list1, str(strings[item]));
      strings[item] = strings[--nr_strings]; 
    }

  *list2 = list_create((void_cmp) strcmp);
  for (size_t input_size = 4 + (rand() % 4); input_size; --input_size)
    {
      int item = rand() % nr_strings; 
      list_insert(*list2, str(strings[item]));
      strings[item] = strings[--nr_strings]; 
    }
}

void print(list_t *list)
{
    char *str = list_print(list, NULL);
    printf("%s\n", str);
    free(str);
}

bool link_check(link_t *links[], int size, link_t *first)
{
 for (int i = 0; i < size; ++i)
    {
      link_t *cursor = first;
      while (cursor)
        {
          if (cursor == links[i]) break;
          cursor = cursor->next;
        }
      if (cursor == NULL)
        {
          printf("Missing link with element %s\n", (char *)links[i]->element);
          return false; 
        }
    }
  return true;
}

int main(void)
{
  for (int i = 0; i < 3; ++i)
    {
      puts("=========================== TEST START =================================");
      list_t *a = NULL;
      list_t *b = NULL;
      int exp_size = 0;
      printf("Expected size of 'b' = %d. Returned size = %d.",exp_size,list_size(b));
      printf("  . . .  %s\n",(list_size(b) == exp_size?"PASSED":"FAILED!"));
      create_random_lists(&a, &b);

      printf(" First random list 'a':  "); print(a);
      printf(" Returned size of 'a':    %d\n",list_size(a));
      printf("Second random list 'b':  "); print(b);
      printf(" Returned size of 'b':    %d\n",list_size(b));

      puts("Merging the lists.");
      list_merge(a, b);

      /// Recording pointer values:
      int size_a = list_size(a);
      int size_b = list_size(b);
      int size = size_a + size_b;
      link_t *links[size];
      int i = 0;
      for (link_t *cursor = list_first(a); cursor; cursor = cursor->next)
        {
          links[i++] = cursor; 
        }
      for (link_t *cursor = list_first(b); cursor; cursor = cursor->next)
        {
          links[i++] = cursor; 
        }

      printf("Expecting size of 'a' to be zero: ");
      printf("%d == %d ... %s\n", list_size(a), 0, (list_size(a) == 0 ? "PASSED" : "FAILED!!"));
      printf("Expecting size of 'b' to be sum of 'a' and 'b' before merge: ");
      printf("%d == %d ... %s\n", list_size(b), size, (list_size(b) == size ? "PASSED" : "FAILED!!!!"));

      printf("Expecting the output of 'a' to be empty:  "); print(a);
      puts("Expecting the output of 'b' to be all elements of 'a' and 'b' before merge");
      puts(" - in ascending order - e.g. a < b, b < c, etc.:");
      print(b);

      printf("Checking that the links, not just elements were moved");
      printf("... %s\n", link_check(links, size, list_first(b)) ? "PASSED" : "FAILED");
      
      list_destroy(a);
      //if (a != NULL) puts("a - BLÄÄÄÄH!");
      list_destroy(b);
      //if (b != NULL) puts("b - BLÄÄÄÄH!");
      puts("=========================== TEST STOP ==================================\n");
    }
  
  return 0;
}
