// Imperative and Object-Oriented Programming Methodology
// Uppsala University - 2018 Autumn
// Copyright © 2018 Henrik Schulze & Tobias Wrigstad
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include "yourcode.h"

struct list
{
  link_t *first;
  link_t *last;
  void_cmp cmp;
};

static link_t *link_create(link_t *next, void *element)
{
  link_t *link_created = calloc(1,sizeof(link_t));
  link_created->next = next;
  link_created->element = element;
  return link_created;
}

list_t *list_create(void_cmp cmp)
{
  list_t *list = calloc(1,sizeof(list_t));
  list->first = list->last = link_create(NULL, NULL); /// Dummy!
  list->cmp = cmp;
  return list;
}

void list_insert(list_t *list, void *element)
{
  link_t *cursor = list->first;
  while (cursor->next != NULL && list->cmp(cursor->next->element, element) < 0)
    {
      cursor = cursor->next;
    }
  if (cursor->next != NULL)
    {
      if (list->cmp(cursor->next->element, element) > 0)
        {
          cursor->next = link_create(cursor->next, element);
        }
      //////else // Do nothing! - The element is already in the list.
        //////{
          //////cursor->next->element = element;
        //////}
    }
  else
    {
      list->last = cursor->next = link_create(NULL, element);
    }
}

/// Returnera antalet element i listan. Detta skall räknas ut
/// genom iteration över samtliga länkar.
int list_size(list_t *list)
{
  /// TODO
  int list_size = 0;
  if (list == NULL || list->first == NULL) { return list_size; }
  link_t *cursor = list->first;
  while (cursor->next != NULL)
    {
      ++list_size;
      cursor = cursor->next;
    }
  return list_size;
}

/// Ta bort listan, alla länkar och alla länkars alla element
void list_destroy(list_t *list)
{
  /// TODO
  if (list == NULL) { return; }
  link_t *cursor = list->first;
  while (cursor != NULL)
    {
      link_t *destroy_me = cursor;
      cursor = cursor->next;
      free(destroy_me->element);
      free(destroy_me);
    }
  free(list);
  list = NULL;
}

/// Helper function for removing a link from the source list and
/// return it so that it can be added to the destination list.
static link_t *link_insert(link_t* dest_next, link_t* source)
{
  if (source == NULL || source->next == NULL)
    {
      puts(" UPS! - ERROR entering link_insert()!");
      return NULL;
    }
  link_t* return_source = source->next;
  source->next = source->next->next;// This removes return_source from the source list.
  return_source->next = dest_next;  // Prepare for inserting into the destination list.
  return return_source;
}

/// Alla länkar (och dess medföljande element, dock ej dummies)
/// skall flyttas från source till dest.
void list_merge(list_t *source, list_t *dest)
{
  /// TODO
  if (source == NULL || source->first == NULL || source->first->next == NULL)
    {   // The source list is empty. There is nothing to move:
      return;
    }
  link_t *scursor = source->first; // Coming here, scursor->next != NULL, initially.
  link_t *dcursor = dest->first;
  while (scursor->next != NULL)
    {
      while (dcursor->next != NULL && dest->cmp(dcursor->next->element, scursor->next->element) < 0)
        {
          dcursor = dcursor->next;
        }
      if (dcursor->next != NULL)
        {
          //if (dest->cmp(dcursor->next->element, scursor->next->element) >= 0)
            //{
              dcursor->next = link_insert(dcursor->next, scursor);
            //}
        }
      else
        {
          dest->last = dcursor->next = link_insert(dcursor->next, scursor);
        }
    }
  scursor = scursor->next;
}

/// Returns a string representation of the list to be printed
/// No need to read or change this function!!
char *list_print(list_t *list, elem_to_string to_string)
{
  char *result = calloc(2048, sizeof(char));
  strcat(result, "[");
  
  for (link_t *cursor = list->first->next; cursor != NULL; cursor = cursor->next)
    {
      char *tmp = NULL;
      asprintf(&tmp,"%s, ",to_string?
        to_string(cursor->element):(char*)cursor->element);
      strcat(result, tmp);
      free(tmp);
    }

  int size = strlen(result);
  strcpy(result + (size > 1 ? size - 2 : size), "]");
  
  return result;
}

/// Return the first link in the list for testing
/// No need to read or change this function!!
link_t *list_first(list_t *list)
{
  return list->first->next;
}
